# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Top-level package for Utility module."""

from .content_util import contents_key
from .content_util import format_key
from .expression_util import build_query_expression_for_commit_log_flags
from .expression_util import build_query_expression_for_contents_listing_flags
from .expression_util import parse_to_iso8601

__all__ = [
    "format_key",
    "contents_key",
    "build_query_expression_for_commit_log_flags",
    "build_query_expression_for_contents_listing_flags",
    "parse_to_iso8601",
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Configure pytest."""

import os
import shutil
import tempfile
from typing import List
from typing import Optional

import pytest
from _pytest.fixtures import FixtureRequest
from assertpy import assert_that
from click.testing import CliRunner
from click.testing import Result
from pytest import Session
from vcr.request import Request

from pynessie import cli
from pynessie.model import ReferenceSchema

# Per-test Nessie config directly
nessie_config_dir: str

# Helper variable to control VCR recording in test fixtures
nessie_cleanup: bool = False

def pytest_configure(config):  # noqa
    """Configure pytest."""
    config.addinivalue_line("markers", "doc: mark as end-to-end test.")


def pytest_sessionstart(session: "Session") -> None:
    """Setup a fresh temporary config directory for tests."""
    global nessie_config_dir
    nessie_config_dir = tempfile.mkdtemp() + "/"
    # Instruct Confuse to keep Nessie config file in the temp location:
    os.environ["NESSIEDIR"] = nessie_config_dir


def pytest_sessionfinish(session: "Session", exitstatus: int) -> None:
    """Remove temporary config directory."""
    global nessie_config_dir
    shutil.rmtree(nessie_config_dir)


def _run(args: List[str], input: Optional[str] = None, ret_val: int = 0) -> Result:
    result = CliRunner().invoke(cli.cli, args, input=input)
    if result.exit_code != ret_val:
        print(result.output)
        print(result)
    assert_that(result.exit_code).is_equal_to(ret_val)
    return result


def _cli(args: List[str], input: Optional[str] = None, ret_val: int = 0) -> str:
    return _run(args, input, ret_val).output


def reset_nessie_server_state() -> None:
    """Resets the Nessie Server to an initial, clean state for testing."""
    # Delete all branches
    branches = ReferenceSchema().loads(_cli(["--json", "branch"]), many=True)
    for branch in branches:
        _cli(["branch", "-d", branch.name])

    # Delete all tags
    tags = ReferenceSchema().loads(_cli(["--json", "tag"]), many=True)
    for tag in tags:
        _cli(["tag", "-d", tag.name])

    # Note: This hash should match the java constant AbstractDatabaseAdapter.NO_ANCESTOR
    no_ancestor_hash = "2e1cfa82b035c26cbbbdae632cea070514eb8b773f616aaeaf668e2f0be8f10d"

    # Re-create the main branch from the "root" (a.k.a. no ancestor) hash
    _cli(["branch", "--force", "-o", no_ancestor_hash, "main", "main"])

    # Verify the re-created main branch
    branches = ReferenceSchema().loads(_cli(["--json", "branch"]), many=True)
    assert_that(branches).is_length(1)
    assert_that(branches[0].name).is_equal_to("main")
    assert_that(branches[0].hash_).is_equal_to(no_ancestor_hash)


def before_record_cb(request: Request) -> Optional[Request]:
    """VCR callback that instructs it to not record our "cleanup" requests."""
    if nessie_cleanup:
        return None
    return request


@pytest.fixture(scope="module")
def vcr_config() -> dict:
    """VCR config that adds a custom before_record_request callback."""
    pytest.nessie_cleanup = False
    return {
        "before_record_request": before_record_cb,
    }


@pytest.fixture(autouse=True)
def _clean_nessie_session_marker(request: "FixtureRequest", record_mode: str) -> None:
    """This pytest fixture is invoked for all test methods and cleans up the Nessie Server state.

    :param request Request object provided by the pytest framework
    :param record_mode Recording mode provided by the VCR plugin
    """
    # Cleaning the Nessie Server state is meaningful only when we are recording,
    # i.e. when the tests are running against a real server, not against VCR cassettes.
    # Consequently, when the tests will run against VCR cassettes, the pre-recorded
    # responses will simulate a "clean" server.
    if record_mode is None or record_mode == "none":
        return

    # Clean the server state for tests that have the @pytest.mark.clean_nessie_session annotation
    if request.node.get_closest_marker("vcr"):
        global nessie_cleanup
        try:
            nessie_cleanup = True
            reset_nessie_server_state()
        finally:
            nessie_cleanup = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tests for `pynessie` package."""
import itertools
import os
from pathlib import Path

import confuse
import pytest
import simplejson
from assertpy import assert_that

from pynessie import __version__
from pynessie.model import Branch
from pynessie.model import Contents
from pynessie.model import ContentsSchema
from pynessie.model import DeltaLakeTable
from pynessie.model import EntrySchema
from pynessie.model import IcebergTable
from pynessie.model import ReferenceSchema
from pynessie.model import SqlView
from .conftest import _cli
from .conftest import _run


@pytest.mark.vcr
def test_command_line_interface() -> None:
    """Test the CLI."""
    assert "Usage: cli" in _cli(list())
    assert "Usage: cli" in _cli(["--help"])
    assert __version__ in _cli(["--version"])
    references = ReferenceSchema().loads(_cli(["--json", "branch", "-l"]), many=True)
    assert len(references) == 1
    assert references[0].name == "main"
    assert isinstance(references[0], Branch)


def test_config_options() -> None:
    """Ensure config cli option is consistent."""
    assert "Usage: cli" in _cli(["config"])
    vars = ["--add x", "--get x", "--list", "--unset x"]
    for i in itertools.permutations(vars, 2):
        assert "Error: Illegal usage: " in _cli(["config"] + [*i[0].split(" "), *i[1].split(" ")], ret_val=2)

    _cli(["config", "x", "--add", "x"])


def test_set_unset() -> None:
    """Test config set/unset/list."""
    _cli(["config", "--add", "test.data", "123", "--type", "int"])
    assert _cli(["config", "test.data", "--type", "int"]) == "123\n"
    _cli(["config", "--unset", "test.data"])
    assert "123" not in _cli(["config", "--list"])


@pytest.mark.vcr
def test_remote() -> None:
    """Test setting and viewing remote."""
    _cli(["remote", "add", "http://test.url"])
    _cli(["remote", "add", "http://localhost:19120/api/v1"])
    assert "main" in _cli(["--json", "remote", "show"])
    _cli(["remote", "set-head", "dev"])
    assert _cli(["config", "default_branch"]) == "dev\n"
    _cli(["remote", "set-head", "dev", "-d"])
    result = _run(["config", "default_branch"], ret_val=1)
    assert result.output == ""
    assert isinstance(result.exception, confuse.exceptions.ConfigTypeError)


def _new_table(table_id: str) -> IcebergTable:
    return IcebergTable(table_id, "/a/b/c", "xyz")


def _make_commit(
    key: str, table: Contents, branch: str, head_hash: str = None, message: str = "test message", author: str = "nessie test"
) -> None:
    if not head_hash:
        refs = {i.name: i.hash_ for i in ReferenceSchema().loads(_cli(["--json", "branch"]), many=True)}
        head_hash = refs[branch]
    _cli(
        ["contents", "--set", "--stdin", key, "--ref", branch, "-m", message, "-c", head_hash, "--author", author],
        input=ContentsSchema().dumps(table),
    )


@pytest.mark.vcr
def test_log() -> None:
    """Test log and log filtering."""
    logs = simplejson.loads(_cli(["--json", "log"]))
    assert len(logs) == 0
    _cli(["branch", "dev_test_log"])
    table = _new_table("test_log_dev")
    _make_commit("log.foo.dev", table, "dev_test_log", author="nessie_user1")
    table = _new_table("test_log")
    _make_commit("log.foo.bar", table, "main", author="nessie_user1")
    tables = ContentsSchema().loads(_cli(["--json", "contents", "log.foo.bar"]), many=True)
    assert len(tables) == 1
    assert tables[0] == table
    logs = simplejson.loads(_cli(["--json", "log"]))
    assert len(logs) == 1
    logs = simplejson.loads(_cli(["--json", "log", "--revision-range", logs[0]["hash"]]))
    assert len(logs) == 1
    entries = EntrySchema().loads(_cli(["--json", "contents", "--list"]), many=True)
    assert len(entries) == 1
    _cli(
        [
            "--json",
            "contents",
            "--delete",
            "log.foo.bar",
            "--ref",
            "main",
            "-m",
            "delete_message",
            "-c",
            logs[0]["hash"],
            "--author",
            "nessie_user2",
        ],
    )
    logs = simplejson.loads(_cli(["--json", "log", "-n", 1]))
    assert len(logs) == 1
    logs = simplejson.loads(_cli(["--json", "log", "dev_test_log"]))
    assert len(logs) == 1
    logs = simplejson.loads(_cli(["--json", "log"]))
    assert len(logs) == 2
    logs = simplejson.loads(_cli(["--json", "log", "--revision-range", "{}..{}".format(logs[0]["hash"], logs[1]["hash"])]))
    assert len(logs) == 1
    logs = simplejson.loads(_cli(["--json", "log"]))
    assert len(logs) == 2
    logs = simplejson.loads(_cli(["--json", "log", "--author", "nessie_user1"]))
    assert len(logs) == 1
    assert_that(logs[0]["author"]).is_equal_to("nessie_user1")
    logs = simplejson.loads(_cli(["--json", "log", "--author", "nessie_user2"]))
    assert len(logs) == 1
    assert_that(logs[0]["author"]).is_equal_to("nessie_user2")
    logs = simplejson.loads(_cli(["--json", "log", "--author", "nessie_user2", "--author", "nessie_user1"]))
    assert len(logs) == 2
    # the committer is set on the server-side and is empty if we're not logged
    # in when performing a commit
    logs = simplejson.loads(_cli(["--json", "log", "--committer", ""]))
    assert len(logs) == 2
    logs = simplejson.loads(_cli(["--json", "log", "--query", "commit.author == 'nessie_user2' || commit.author == 'non_existing'"]))
    assert len(logs) == 1
    logs = simplejson.loads(_cli(["--json", "log", "--after", "2001-01-01T00:00:00+00:00", "--before", "2999-12-30T23:00:00+00:00"]))
    assert len(logs) == 2


@pytest.mark.vcr
def test_branch() -> None:
    """Test create and assign refs."""
    references = ReferenceSchema().loads(_cli(["--json", "branch"]), many=True)
    assert len(references) == 1
    _cli(["branch", "dev"])
    references = ReferenceSchema().loads(_cli(["--json", "branch"]), many=True)
    assert len(references) == 2
    _cli(["branch", "etl", "main"])
    references = ReferenceSchema().loads(_cli(["--json", "branch"]), many=True)
    assert len(references) == 3
    references = ReferenceSchema().loads(_cli(["--json", "branch", "-l", "etl"]), many=False)
    assert_that(references.name).is_equal_to("etl")
    references = simplejson.loads(_cli(["--json", "branch", "-l", "foo"]))
    assert len(references) == 0
    _cli(["branch", "-d", "etl"])
    _cli(["branch", "-d", "dev"])
    references = ReferenceSchema().loads(_cli(["--json", "branch"]), many=True)
    assert len(references) == 1


@pytest.mark.vcr
def test_tag() -> None:
    """Test create and assign refs."""
    references = ReferenceSchema().loads(_cli(["--json", "tag"]), many=True)
    assert len(references) == 0
    _cli(["tag", "dev-tag", "main"])
    references = ReferenceSchema().loads(_cli(["--json", "tag"]), many=True)
    assert len(references) == 1
    _cli(["tag", "etl-tag", "main"])
    references = ReferenceSchema().loads(_cli(["--json", "tag"]), many=True)
    assert len(references) == 2
    references = ReferenceSchema().loads(_cli(["--json", "tag", "-l", "etl-tag"]), many=False)
    assert_that(references.name).is_equal_to("etl-tag")
    references = simplejson.loads(_cli(["--json", "tag", "-l", "foo"]))
    assert len(references) == 0
    _cli(["tag", "-d", "etl-tag"])
    _cli(["tag", "-d", "dev-tag"])
    references = ReferenceSchema().loads(_cli(["--json", "tag"]), many=True)
    assert len(references) == 0
    _cli(["tag", "v1.0"])
    tags = {i.name: i.hash_ for i in ReferenceSchema().loads(_cli(["--json", "tag"]), many=True)}
    branches = {i.name: i.hash_ for i in ReferenceSchema().loads(_cli(["--json", "branch"]), many=True)}
    assert tags["v1.0"] == branches["main"]


@pytest.mark.vcr
def test_commit_with_expected_state() -> None:
    """Test making a commit with some expected Contents, i.e. IcebergTable."""
    _cli(["branch", "dev"])
    _make_commit("commit.expected.contents", _new_table("test_expected_contents"), "dev", message="commit 1")
    # the second commit will use the Contents of the first one as "expected contents"
    _make_commit("commit.expected.contents", _new_table("test_expected_contents"), "dev", message="commit 2")


@pytest.mark.vcr
def test_commit_no_expected_state() -> None:
    """Test making two commit without any expected Contents, i.e. using DeltaLakeTable."""
    _cli(["branch", "dev"])
    table1 = DeltaLakeTable(
        id="test_commit_no_expected_state", metadata_location_history=["asd111"], checkpoint_location_history=["def"], last_checkpoint="x"
    )
    _make_commit("commit.expected.contents", table1, "dev", message="commit 1")
    table2 = DeltaLakeTable(
        id="test_commit_no_expected_state", metadata_location_history=["asd222"], checkpoint_location_history=["def"], last_checkpoint="x"
    )
    # the second commit will set new contents without the "expected contents" parameter due to using a DeltaLakeTable
    _make_commit("commit.expected.contents", table2, "dev", message="commit 2")


@pytest.mark.vcr
def test_assign() -> None:
    """Test assign operation."""
    _cli(["branch", "dev"])
    _make_commit("assign.foo.bar", _new_table("test_assign"), "dev")
    _run(["branch", "main", "dev", "--force"])
    branches = ReferenceSchema().loads(_cli(["--json", "branch"]), many=True)
    refs = {i.name: i.hash_ for i in branches}
    assert refs["main"] == refs["dev"]
    _cli(["tag", "v1.0", "main"])
    tags = {i.name: i.hash_ for i in ReferenceSchema().loads(_cli(["--json", "tag"]), many=True)}
    assert tags["v1.0"] == refs["main"]
    _cli(["tag", "v1.0", "dev", "--force"])
    tags = {i.name: i.hash_ for i in ReferenceSchema().loads(_cli(["--json", "tag"]), many=True)}
    assert tags["v1.0"] == refs["dev"]


@pytest.mark.vcr
def test_merge() -> None:
    """Test merge operation."""
    _cli(["branch", "dev"])
    _make_commit("merge.foo.bar", _new_table("test_merge"), "dev")
    ref = ReferenceSchema().loads(_cli(["--json", "branch", "-l", "main"]), many=False)
    main_hash = ref.hash_
    _cli(["merge", "dev", "-c", main_hash])
    branches = ReferenceSchema().loads(_cli(["--json", "branch"]), many=True)
    refs = {i.name: i.hash_ for i in branches}
    assert refs["main"] == refs["dev"]


@pytest.mark.vcr
def test_transplant() -> None:
    """Test transplant operation."""
    _cli(["branch", "dev"])
    _make_commit("transplant.foo.bar", _new_table("test_transplant_1"), "dev")
    _make_commit("bar.bar", _new_table("test_transplant_2"), "dev")
    _make_commit("foo.baz", _new_table("test_transplant_3"), "dev")
    refs = ReferenceSchema().loads(_cli(["--json", "branch", "-l"]), many=True)
    main_hash = next(i.hash_ for i in refs if i.name == "main")
    logs = simplejson.loads(_cli(["--json", "log", "dev"]))
    first_hash = [i["hash"] for i in logs]
    _cli(["cherry-pick", "-c", main_hash, "-s", "dev", first_hash[1], first_hash[0]])

    logs = simplejson.loads(_cli(["--json", "log"]))
    assert len(logs) == 2  # two commits were transplanted into an empty `main`


@pytest.mark.doc
def test_all_help_options() -> None:
    """Write out all help options to std out."""
    args = ["", "config", "branch", "tag", "remote", "log", "merge", "cherry-pick", "contents"]

    for i in args:
        result = _cli([x for x in [i] if x] + ["--help"])
        cwd = os.getcwd()
        with open(Path(Path(cwd), "docs", "{}.rst".format(i if i else "main")), "w") as f:
            f.write(".. code-block:: bash\n\n\t")
            for line in result.split("\n"):
                f.write(line + "\n\t")
            f.write("\n\n")


@pytest.mark.vcr
def test_contents_listing() -> None:
    """Test contents listing and filtering."""
    branch = "contents_listing_dev"
    _cli(["branch", branch])

    iceberg_table = IcebergTable(id="test_contents_listing", metadata_location="/a/b/c", id_generators="xyz")
    delta_lake_table = DeltaLakeTable(
        id="uuid2", metadata_location_history=["asd"], checkpoint_location_history=["def"], last_checkpoint="x"
    )
    sql_view = SqlView(id="uuid3", sql_text="SELECT * FROM foo", dialect="SPARK")

    _make_commit("this.is.iceberg.foo", iceberg_table, branch)
    _make_commit("this.is.delta.bar", delta_lake_table, branch)
    _make_commit("this.is.sql.baz", sql_view, branch)

    tables = ContentsSchema().loads(_cli(["--json", "contents", "--ref", branch, "this.is.iceberg.foo"]), many=True)
    assert_that(tables).is_length(1)
    assert_that(tables[0]).is_equal_to(iceberg_table)

    tables = ContentsSchema().loads(_cli(["--json", "contents", "--ref", branch, "this.is.delta.bar"]), many=True)
    assert_that(tables).is_length(1)
    assert_that(tables[0]).is_equal_to(delta_lake_table)

    tables = EntrySchema().loads(_cli(["--json", "contents", "--ref", branch, "--list", "--type", "ICEBERG_TABLE"]), many=True)
    assert_that(tables).is_length(1)
    assert_that(tables[0].kind).is_equal_to("ICEBERG_TABLE")

    tables = EntrySchema().loads(_cli(["--json", "contents", "--ref", branch, "--list", "--type", "DELTA_LAKE_TABLE"]), many=True)
    assert_that(tables).is_length(1)
    assert_that(tables[0].kind).is_equal_to("DELTA_LAKE_TABLE")

    result = _cli(["--json", "contents", "--ref", branch, "--list", "--query", "entry.contentType == 'ICEBERG_TABLE'"])
    tables = EntrySchema().loads(result, many=True)
    assert_that(tables).is_length(1)
    assert_that(tables[0].kind).is_equal_to("ICEBERG_TABLE")

    result = _cli(
        ["--json", "contents", "--ref", branch, "--list", "--query", "entry.contentType in ['ICEBERG_TABLE', 'DELTA_LAKE_TABLE']"]
    )
    tables = EntrySchema().loads(result, many=True)
    assert_that(tables).is_length(2)
    assert_that(set(t.kind for t in tables)).is_equal_to({"DELTA_LAKE_TABLE", "ICEBERG_TABLE"})

    result = _cli(["--json", "contents", "--ref", branch, "--list", "--query", "entry.namespace.startsWith('this.is.del')"])
    tables = EntrySchema().loads(result, many=True)
    assert_that(tables).is_length(1)
    assert_that(tables[0].kind).is_equal_to("DELTA_LAKE_TABLE")

    result = _cli(["--json", "contents", "--ref", branch, "--list", "--query", "entry.namespace.startsWith('this.is')"])
    tables = EntrySchema().loads(result, many=True)
    assert_that(tables).is_length(3)
    assert_that(set(i.kind for i in tables)).is_equal_to({"ICEBERG_TABLE", "VIEW", "DELTA_LAKE_TABLE"})



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Authentication tests for Nessi CLI."""
import pytest

from .conftest import _cli


@pytest.fixture(scope="module")
def vcr_config() -> dict:
    """Configures VCR to also match on HTTP headers."""
    return {"match_on": ["method", "scheme", "host", "port", "path", "query", "headers"]}


@pytest.mark.vcr
def test_bearer_auth() -> None:
    """Test the "remote show" command with bearer authentication."""
    assert "main" in _cli(["--json", "--auth-token", "test_token_123", "remote", "show"])


@pytest.mark.vcr
def test_bearer_auth_invalid() -> None:
    """Test any command with an invalid bearer authentication token."""
    result = _cli(["--json", "--auth-token", "invalid", "remote", "show"], ret_val=1)
    assert "Client Error" in result
    assert "Unauthorized" in result



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Utility functions that can be used to manipulate contents like keys and values."""

import re

from ..model import ContentsKey

_dot_regex = re.compile('\\.(?=([^"]*"[^"]*")*[^"]*$)')


def format_key(raw_key: str) -> str:
    """Format given key based on regex."""
    elements = _dot_regex.split(raw_key)
    return ".".join(i.replace(".", "\0") for i in elements if i)


def contents_key(raw_key: str) -> ContentsKey:
    """Format given content key based on regex."""
    elements = _dot_regex.split(raw_key)
    return ContentsKey([i for i in elements if i])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""contents CLI command."""

from collections import defaultdict
from typing import Any, Generator, Optional
from typing import List

import click

from ..cli_common_context import ContextObject, MutuallyExclusiveOption, pass_client
from ..client import NessieClient
from ..error import error_handler, NessieNotFoundException
from ..model import Contents, ContentsSchema, Delete, Entries, Entry, EntrySchema, Operation, Put
from ..utils import build_query_expression_for_contents_listing_flags, contents_key, format_key


@click.command("contents")
@click.option(
    "-l",
    "--list",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="list tables",
    mutually_exclusive=["delete", "set"],
)
@click.option(
    "-d",
    "--delete",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="delete a table",
    mutually_exclusive=["list", "set"],
)
@click.option(
    "-s",
    "--set",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="modify a table",
    mutually_exclusive=["list", "delete"],
)
@click.option(
    "-i",
    "--stdin",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="read contents for --set from STDIN (separated by Ctrl-D)",
    mutually_exclusive=["list", "delete"],
)
@click.option(
    "-S",
    "--expect-same-contents",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="send the same contents both as the new and expected (old contents) parameters for --set operations",
    mutually_exclusive=["list", "delete"],
)
@click.option(
    "-c",
    "--condition",
    "expected_hash",
    help="Expected hash. Only perform the action if the branch currently points to the hash specified by this option.",
)
@click.option("-r", "--ref", help="branch to list from. If not supplied the default branch from config is used")
@click.option("-m", "--message", help="commit message")
@click.option(
    "-t",
    "--type",
    "entity_types",
    help="entity types to filter on, if no entity types are passed then all types are returned",
    multiple=True,
)
@click.option(
    "--query",
    "--query-expression",
    "query_expression",
    multiple=False,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["entity_type"],
    help="Allows advanced filtering using the Common Expression Language (CEL). "
    "An intro to CEL can be found at https://github.com/google/cel-spec/blob/master/doc/intro.md.\n"
    "Some examples with usable variables 'entry.namespace' (string) & 'entry.contentType' (string) are:\n"
    "entry.namespace.startsWith('a.b.c')\n"
    "entry.contentType in ['ICEBERG_TABLE','DELTA_LAKE_TABLE']\n"
    "entry.namespace.startsWith('some.name.space') && entry.contentType in ['ICEBERG_TABLE','DELTA_LAKE_TABLE']\n",
)
@click.option("--author", help="The author to use for the commit")
@click.argument("key", nargs=-1, required=False)
@pass_client
@error_handler
def contents(
    ctx: ContextObject,
    list: bool,
    delete: bool,
    set: bool,
    stdin: bool,
    expect_same_contents: bool,
    key: List[str],
    ref: str,
    message: str,
    expected_hash: str,
    entity_types: List[str],
    author: str,
    query_expression: str,
) -> None:
    """Contents operations.

    KEY name of object to view, delete. If listing the key will limit by namespace what is included.
    """
    if list:
        keys = ctx.nessie.list_keys(
            ref if ref else ctx.nessie.get_default_branch(),
            query_expression=build_query_expression_for_contents_listing_flags(query_expression, entity_types),
        )
        results = EntrySchema().dumps(_format_keys_json(keys, *key), many=True) if ctx.json else _format_keys(keys, *key)
    elif delete or set:
        ops = _get_contents(ctx.nessie, ref, delete, stdin, expect_same_contents, *key)
        if ops:
            ctx.nessie.commit(ref, expected_hash, _get_message(message), author, *ops)
        else:
            click.echo("No changes requested")
        results = ""
    else:

        def content(*x: str) -> Generator[Contents, Any, None]:
            return ctx.nessie.get_values(ref if ref else ctx.nessie.get_default_branch(), *x)

        results = ContentsSchema().dumps(content(*key), many=True) if ctx.json else "\n".join((i.pretty_print() for i in content(*key)))
    click.echo(results)


def _edit_contents(key: str, body: str) -> Optional[str]:
    marker = "# Everything below is ignored\n"
    edit_message = (
        body
        + "\n\n"
        + marker
        + "Edit the content above to commit changes."
        + " Closing without change will result in a no-op."
        + " Removing the content results in a delete"
        + "\nContents Key: "
        + key
    )
    try:
        message = click.edit(edit_message)
        # Note: None can occur when the user quits the editor without saving
        if message is None:
            return None
        message_altered = message.split(marker, 1)[0].strip("\n")
        return message_altered
    except click.ClickException:
        pass
    return None


def _get_contents(
    nessie: NessieClient, ref: str, delete: bool = False, use_stdin: bool = False, expect_same_contents: bool = False, *keys: str
) -> List[Operation]:
    contents_altered: List[Operation] = list()
    for raw_key in keys:
        key = format_key(raw_key)
        content_orig = None

        try:  # get current contents, if present
            content = nessie.get_values(ref, key)
            content_orig = content.__next__()
        except NessieNotFoundException:
            pass

        if delete:
            content_json = None
        elif use_stdin:
            click.echo("Enter JSON contents for key " + key)
            content_json = click.get_text_stream("stdin").read()
            content_json = content_json.strip("\n")
        else:  # allow the user to provide interactive input via the shell $EDITOR
            content_json = _edit_contents(key, (ContentsSchema().dumps(content_orig) if content_orig else ""))
            if content_json is None:
                click.echo("Skipping key " + key + " (contents not edited)")
                continue

        if content_json:
            click.echo("Setting contents for key " + key)
            content_new = ContentsSchema().loads(content_json)
            if content_new.requires_expected_state():
                content_expected = content_new if expect_same_contents else content_orig
                contents_altered.append(Put(contents_key(raw_key), content_new, content_expected))
            else:
                contents_altered.append(Put(contents_key(raw_key), content_new))
        else:
            click.echo("Deleting contents for key " + key)
            contents_altered.append(Delete(contents_key(raw_key)))

    return contents_altered


def _get_commit_message(*keys: str) -> Optional[str]:
    MARKER = "# Everything below is ignored\n"
    message = click.edit("\n\n" + MARKER + "Edit the commit message above." + "\n".join(keys))
    return message.split(MARKER, 1)[0].rstrip("\n") if message is not None else None


def _get_message(message: str, *keys: str) -> Optional[str]:
    if message:
        return message
    return _get_commit_message(*keys)


def _format_keys_json(keys: Entries, *expected_keys: str) -> List[Entry]:
    results = list()
    for k in keys.entries:
        value = ['"{}"'.format(i) if "." in i else i for i in k.name.elements]
        if expected_keys and all(key for key in expected_keys if key not in value):
            continue
        results.append(k)
    return results


def _format_keys(keys: Entries, *key: str) -> str:
    results = defaultdict(list)
    result_str = ""
    for e in keys.entries:
        results[e.kind].append(e.name)
    for k in results.keys():
        result_str += k + ":\n"
        for v in results[k]:
            value = ['"{}"'.format(i) if "." in i else i for i in v.elements]
            if key and all(k for k in key if k not in value):
                continue
            result_str += "\t{}\n".format(".".join(value))
    return result_str
